#include "Wire.h"  // Only needed for Arduino 1.6.5 and earlier
#include "SSD1306.h" // alias for `#include "SSD1306Wire.h"`

#include "SPI.h"
#include "WiFi.h"
#include "LoRa.h"
#include "PubSubClient.h"
#include "CRC32.h"

#include "TimeClient.h"
#include "Ticker.h";

#include "SketchFonts.h"
#include "SketchImages.h"

#include "LangloLoRa.h"       // LoRa configuration parameters
#include "LangloWiFi.h"       // WiFi configuration parameters
#include "LangloMQTT.h"       // MQTT configuration parameters
#include "LangloPackets.h"    // Packet structures
#include "LangloNodes.h"      // Local 'hosts' file (not actaully used at the moment)
#include "nodeG1config.h"     // Specific Node configuration parameters

#include <esp_int_wdt.h>
#include <esp_task_wdt.h>

const float batteryVoltageFloor = 0.0;
const float batteryVoltageCeiling = 12.0;

const uint8_t tankLevelFloor = 0;
const uint8_t tankLevelCeiling = 100;

const float rainfallFloor = 0.0;
const float rainfallCeiling = 500.0;

const float temperatureFloor = -100.0;
const float temperatureCeiling = 100.0;

const uint16_t pressureFloor = 800;
const uint16_t pressureCeiling = 1100;

const uint16_t humidityFloor = 0;
const uint16_t humidityCeiling = 100;

const uint16_t windDirectionFloor = 0;
const uint16_t windDirectionCeiling = 360;

const uint8_t windSpeedFloor = 0;
const uint8_t windSpeedCeiling = 250;

// Heartbeat variables
const int heartbeatInterval = 60;  // 1 minute
const int healthCheckInterval = 300;  // 5 minutes
const int deadBeat = 5;  // Missed heartbeats before declared dead
int node1Timer = 0;
int node2Timer = 0;
int node3Timer = 0;
int node4Timer = 0;
int node5Timer = 0;
int node6Timer = 0;
int node7Timer = 0;
int node8Timer = 0;
int node9Timer = 0;
int nodeATimer = 0;
int nodeBTimer = 0;

Ticker heartbeatTicker;
Ticker healthCheckTicker;

// use for no hardreset at the fist loop
bool newstart = 0;

CRC32 crc;

//OLED pins to ESP32 GPIOs via this connection:
//OLED_SDA — GPIO4
//OLED_SCL — GPIO15
//OLED_RST — GPIO16

SSD1306 display(0x3c, 4, 15);

// WIFI_LoRa_32 ports

// GPIO5 — SX1278’s SCK
// GPIO19 — SX1278’s MISO
// GPIO27 — SX1278’s MOSI
// GPIO18 — SX1278’s CS
// GPIO14 — SX1278’s RESET
// GPIO26 — SX1278’s IRQ(Interrupt Request)

#define CS 18
#define RST 14
#define DI0 26

// Initialise the MQTT Client
WiFiClient localClient;
PubSubClient client(localClient);

// TimeClient settings
const float utcOffset = 10;
TimeClient timeClient(utcOffset);

dataPacket incomingPacket = {0};
uint32_t sourceMAC = 0;
uint32_t checksum = 0;
uint16_t sequenceNumber = 0;
uint8_t packetType = 0;
String data = "";

// This function is executed when some device publishes a message to a topic to which this MCU is subscribed.
// Change the function below to add logic to your program, so when a device publishes a message to a topic to 
// which this MCU is subscribed you can actually do something.

void callback(String topic, byte* message, unsigned int length) {
  Serial.print("[callback] Message arrived on Topic: ");
  Serial.print(topic);
  Serial.print("  Message: ");
  String messageTemp;
  
  for (int i = 0; i < length; i++) {
    Serial.print((char)message[i]);
    messageTemp += (char)message[i];
  }
  Serial.println();

  // Feel free to add more if statements to control more GPIOs with MQTT

}

void hard_restart() {
  esp_task_wdt_init(1,true);
  esp_task_wdt_add(NULL);
  while(true);
}

void WiFi_Connect() {
  const int retryLimit = 10;
  int retryCount = 0;
  int retryDelay = 0;
  Serial.println();
  while ( retryCount < retryLimit ) {
    if (WiFi.status() == WL_CONNECTED) {
      Serial.println("[WiFi Connect] WiFi Connected");
 
      byte mac[6];
      WiFi.macAddress(mac);
      Serial.print("[WiFi Connect] MAC address: ");
      char dataString[30] = {0};
      sprintf(dataString,"%02X:%02X:%02X:%02X:%02X:%02X",mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
      Serial.println(dataString);
  
      Serial.print("[WiFi Connect] IP address: ");
      Serial.println(WiFi.localIP());
      retryCount = retryLimit;
    } else {
      retryCount = retryCount + 1;
      if (retryCount >= retryLimit) {
        Serial.println( "[WiFi Connect] WiFi connection attempt failed, hard restart..." );
        hard_restart();
      }
      retryDelay = retryDelay + 250;
      WiFi.disconnect();
      delay(retryDelay);
      Serial.println( "[WiFi Connect] Attempting WiFi connection..." );
      WiFi.mode(WIFI_STA);
      delay(retryDelay);
      WiFi.setHostname(myName);
      WiFi.begin(WiFi_SSID, WiFi_Password);
      delay(retryDelay);
    }
  }
}

void MQTT_Connect() {
  const int retryLimit = 10;
  int retryCount = 0;
  Serial.println();
  Serial.println("[MQTT_Connect] Start connection cycle...");
  while (retryCount < retryLimit) {
    retryCount = retryCount + 1;
    // Are we connected?
    if (client.connected()) {
      Serial.println("[MQTT_Connect] Connection complete" );
      // Exit loop next time around
      retryCount = retryLimit;
    } else {
      // Check the WiFi connection
      if (WiFi.status() != WL_CONNECTED) {
        Serial.println("[MQTT_Connect] Attempting WiFi reconnect...");
        WiFi_Connect();
      }
      Serial.println("[MQTT_Connect] Attempting MQTT reconnect...");
      // OK, now try MQTT
      if (client.connect(myName)) {
        Serial.println("[MQTT_Connect] MQTT connected as " + String(myName));
    //      client.subscribe("inTopic");
      } else {
        Serial.print("[MQTT_Connect] MQTT " + String(myName) + " connect failed, rc=");
        Serial.print(client.state());
        Serial.println("[MQTT_Connect] Wait 5 seconds...");      
        delay(5000);
      }
    }
  }
}

void setup() {

  newstart = 1;
  Serial.begin(115200);
  delay(500);
  Serial.println();
  Serial.println("[setup] LoRa [OLED] MQTT Gateway");
  Serial.println("[setup] Commence Set-up...");
  
// Initialise display
  
  Serial.println("[setup] Initialise display...");
  pinMode(16,OUTPUT);
  digitalWrite(16, LOW); // set GPIO16 low to reset OLED
  delay(50);
  digitalWrite(16, HIGH);

  display.init();
  display.clear();
//  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(5,5,"LoRa Receiver");
  display.display();

//  Initialise LoRa

  Serial.println("[setup] Initialise LoRa...");
  SPI.begin(5,19,27,18);
  LoRa.setPins(CS,RST,DI0);
  Serial.println("[setup] Activate LoRa Receiver");
  if (!LoRa.begin(BAND, PABOOST)) {
    display.drawString(5,25,"Starting LoRa failed!");
    Serial.println("[setup] Starting LoRa failed!");
    while (true);
  }
  
  Serial.print("[setup] LoRa Frequency: ");
  Serial.println(BAND);
  
  LoRa.setSpreadingFactor(SpreadingFactor);
  LoRa.setSignalBandwidth(SignalBandwidth);
  LoRa.setPreambleLength(PreambleLength);
  LoRa.setCodingRate4(CodingRateDenominator);
  LoRa.setSyncWord(SyncWord);
  LoRa.disableCrc();
  
  Serial.println("[setup] LoRa Initialisation OK");
  display.drawString(5,25,"LoRa Initialisation OK");
  display.display();

  // Connecting to the WiFi network

  WiFi_Connect();
  client.setServer(MQTT_Server, 1883);
  client.setCallback(callback);

  heartbeatTicker.attach( heartbeatInterval, recordHeartbeat );
  healthCheckTicker.attach( healthCheckInterval, healthCheck );

  Serial.println(); 
  timeClient.updateTime();
  String time = timeClient.getFormattedTime();
  Serial.println(); 
  Serial.println("[setup] Set-up Complete at " + time);
  Serial.println(); 
}

void loop() {

//  Keep in touch...

  if (!client.connected()) {
    Serial.println("[loop] Attempting MQTT (re)connect...");   
    MQTT_Connect();
  }
  
//  Just loop around checking to see if there's anything to process
  
  if ( receivePacket(incomingPacket) ) {
    Serial.println("[loop] Process the packet...");   
    processPacket(incomingPacket);
  }
//  newstart = 0;
}

bool receivePacket(dataPacket& incomingPacket) {
  
// Only if there's something there

  if ( LoRa.parsePacket() ) {
    
//  Serial.println("[receivePacket] Incoming packet: ");
    while ( LoRa.available() ) {
      LoRa.beginPacket();
    
//    Read the packet header
//    The destination address first, to make sure it's for us
    
      for (int i = 0; i < LGO_HeaderSize; i++) {
        incomingPacket.packetByte[i] = LoRa.read();
  /*      Serial.printf( "%02X", incomingPacket.packetByte[i]);
        if ((i % 4) == 3) {
          Serial.println("    ");
        } else {
          Serial.print("  ");
        }
  */
      }
      if (incomingPacket.content.destinationMAC == myMAC) {

//      Read in the rest of the packet
/*        
        Serial.println("[receivePacket] This one's for us...");

        uint32_t sourceMAC = be32toh(incomingPacket.content.sourceMAC);
        int sequenceNumber = (int) be16toh(incomingPacket.content.sequenceNumber);
        uint8_t packetType = incomingPacket.content.type;
*/  
        int byteCount = (int) incomingPacket.content.byteCount;

        Serial.println("[receivePacket] Packet payload...");   

        int i = 0;
        for (i = 0; i < byteCount; i++) {
          incomingPacket.content.payload[i] = LoRa.read();
          Serial.printf( "%02X", incomingPacket.content.payload[i]);
          if ((i % 4) == 3) {
            Serial.println("    ");
          } else {
            Serial.print("  ");
          }
        }
        if ((byteCount % 4) != 0) Serial.println("");
      } else {
        
//      No need to go any further

        Serial.println("[receivePacket] This appears to be for someone else...");
        return false; 
      }
    }
  } else {
    
//  Nothing there

    return false;     
  }
}

bool verifyChecksum(dataPacket& incomingPacket) {
  
  uint32_t packetChecksum = incomingPacket.content.checksum;

//  Calculate the checksum on the received packet

  CRC32 crc;
  int checksumStart = LGO_HeaderSize - 4;
  int checksumEnd = incomingPacket.content.byteCount + 4;
  for (int i = checksumStart; i < checksumEnd; i++) {
    crc.update(incomingPacket.packetByte[i]);
  }
  uint32_t localChecksum = crc.finalize();

  if (packetChecksum == localChecksum) {
    return true;
  } else {
    return false;
  }
}

void processPacket(dataPacket& incomingPacket) {

//  Verify the checksum before we go any furhter

    if (!verifyChecksum(incomingPacket)) {
      Serial.println("[processPacket] Invalid checksum, packet discarded");
      Serial.println();
      return;
    }

// The header first

    sourceMAC = incomingPacket.content.sourceMAC;
      Serial.print("[processPacket] Source MAC: "); 
      Serial.printf("%02X", sourceMAC);
      Serial.println();
    sequenceNumber = (int) incomingPacket.content.sequenceNumber;
      Serial.print("[processPacket] Sequence number: "); 
      Serial.println(sequenceNumber);
    packetType = (int) incomingPacket.content.type;
      Serial.print("[processPacket] Packet Type: ");

    genericPayload packetPayload;
    
    for (int i = 0; i < incomingPacket.content.byteCount; i++) {
      packetPayload.payload[i] = incomingPacket.content.payload[i];
    }

    display.clear();
    display.setFont(ArialMT_Plain_10);

    switch (packetType)
    {
      case powerTypePacket:
      {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(0, 0, "Power");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(128, 0, String(sequenceNumber));
        float batteryVoltage = (float) packetPayload.power.battery.voltage/1000;
        Serial.println("Power");
        Serial.print("  Battery Voltage: ");
        Serial.print(batteryVoltage, 2);
        Serial.println(" V");
        display.setTextAlignment(TEXT_ALIGN_CENTER);
        display.drawString(32, 12, "Battery");
        display.drawString(96, 12, "Panel");
        data = String(batteryVoltage,2);
        data += (" V");
        display.drawString(32,24, data);
        updateMQTT(packetPayload, powerTypePacket, sourceMAC);
        break;
      }
      case voltageTypePacket:
      {
        float sourceVoltage = (float) packetPayload.voltage.source.voltage/1000;
        Serial.println("Voltage: ");
        Serial.print("  Vcc: ");
        Serial.print(sourceVoltage, 2);
        Serial.println(" V");
        updateMQTT(packetPayload, voltageTypePacket, sourceMAC);
        break;
      }
      case tankTypePacket:
      {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(0, 0, "Tank");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(128, 0, String(sequenceNumber));
        uint8_t tankLevel = packetPayload.tank.tank.level;
        Serial.println("Tank");
        Serial.print("  Level: ");
        Serial.print(tankLevel);
        Serial.println(" %");
        display.setTextAlignment(TEXT_ALIGN_CENTER);
        display.drawString(32,12, String(tankLevel) + " %");        
        updateMQTT(packetPayload, tankTypePacket, sourceMAC);
        break;
      }
      case pumpTypePacket:
//      Nothing here yet
        break;
      case weatherTypePacket:
      {
//      Just display atmoshperic stuff on the OLED display for the moment
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(0, 0, "Atmosphere");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(128, 0, String(sequenceNumber));
//        Serial.print("[processPacket] Payload bytes (temperature): ");
//        Serial.printf( "%02X", packetPayload.atmosphere.recorded.temperature.uint32Var );
//        Serial.println();
        uint16_t temperature = packetPayload.weather.recorded.temperature;
        uint16_t pressure = packetPayload.weather.recorded.pressure;
        uint16_t humidity = packetPayload.weather.recorded.humidity;
        Serial.println("Weather Conditions");
        Serial.print("      Temperature: ");
        Serial.print((float) temperature/10, 1);
        Serial.println(" °C");
        Serial.print("         Pressure: ");
        Serial.print(pressure);
        Serial.println(" hPa");
        Serial.print("         Humidity: ");
        Serial.print(humidity);
        Serial.println(" %");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(60, 12, "Temperature");
        display.drawString(60, 24, "Pressure");
        display.drawString(60, 36, "Humidity");
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(64, 12, String((float) temperature/10, 1) + " °C");
        display.drawString(64, 24, String(pressure) + " hPa");
        display.drawString(64, 36, String(humidity) + " %");
        
//      And rainfall and wind details for the serial monitor

        uint16_t recordedRainfall = packetPayload.weather.recorded.rainfall;
        Serial.print("         Rainfall: ");
        Serial.print((float) recordedRainfall/10, 1);
        Serial.println(" mm");
        
        uint16_t windDirection = packetPayload.weather.recorded.windDirection;
        uint8_t windSpeed = packetPayload.weather.recorded.windSpeed;
        Serial.print("   Wind Direction: ");
        Serial.print(windDirection);
        Serial.println(" °");
        Serial.print("       Wind Speed: ");
        Serial.print(windSpeed);
        Serial.println(" kph");
        
        updateMQTT(packetPayload, weatherTypePacket, sourceMAC);
        break;
      }
      case atmosphereTypePacket:
      {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(0, 0, "Atmosphere");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(128, 0, String(sequenceNumber));
//        Serial.print("[processPacket] Payload bytes (temperature): ");
//        Serial.printf( "%02X", packetPayload.atmosphere.recorded.temperature.uint32Var );
//        Serial.println();
        uint16_t temperature = packetPayload.atmosphere.recorded.temperature;
        uint16_t pressure = packetPayload.atmosphere.recorded.pressure;
        uint16_t humidity = packetPayload.atmosphere.recorded.humidity;
        Serial.println("Atmospheric Conditions");
        Serial.print("      Temperature: ");
        Serial.print((float) temperature/10, 1);
        Serial.println(" °C");
        Serial.print("         Pressure: ");
        Serial.print(pressure);
        Serial.println(" hPa");
        Serial.print("         Humidity: ");
        Serial.print(humidity);
        Serial.println(" %");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(60, 12, "Temperature");
        display.drawString(60, 24, "Pressure");
        display.drawString(60, 36, "Humidity");
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(64, 12, String((float) temperature/10, 1) + " °C");
        display.drawString(64, 24, String(pressure) + " hPa");
        display.drawString(64, 36, String(humidity) + " %");
        updateMQTT(packetPayload, atmosphereTypePacket, sourceMAC);
        break;
      }
      case rainfallTypePacket:
      {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(0, 0, "Rainfall");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(128, 0, String(sequenceNumber));
        uint16_t recordedRainfall = packetPayload.rainfall.recorded.rainfall;
        Serial.println("Rainfall");
        Serial.print("  Rainfall: ");
        Serial.print((float) recordedRainfall/10, 1);
        Serial.println(" mm");
        display.setTextAlignment(TEXT_ALIGN_CENTER);
        data = String((float) recordedRainfall/10,1);
        data += (" mm");
        display.drawString(32,12, data);
        updateMQTT(packetPayload, rainfallTypePacket, sourceMAC);
        break;
      }
      case windTypePacket:
      {
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(0, 0, "Wind");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(128, 0, String(sequenceNumber));
//        Serial.print("[processPacket] Payload bytes (windDirection): ");
//        Serial.printf( "%02X", packetPayload.wind.recorded.windDirection );
//        Serial.println();
        uint16_t windDirection = packetPayload.wind.recorded.windDirection;
        uint8_t windSpeed = packetPayload.wind.recorded.windSpeed;
        Serial.println("Wind Conditions");
        Serial.print("      Direction: ");
        Serial.print(windDirection);
        Serial.println(" °");
        Serial.print("          Speed: ");
        Serial.print(windSpeed);
        Serial.println(" kph");
        display.setTextAlignment(TEXT_ALIGN_RIGHT);
        display.drawString(60, 12, "Direction");
        display.drawString(60, 24, "Speed");
        display.setTextAlignment(TEXT_ALIGN_LEFT);
        display.drawString(64, 12, String(windDirection) + " °");
        display.drawString(64, 24, String(windSpeed) + " kph");
        updateMQTT(packetPayload, windTypePacket, sourceMAC);
        break;
      }
      case sprinklerTypePacket:
//      Nothing here yet
        break;
//    default:
//      Nothing here yet
    }
    
    // print RSSI of packet
    Serial.print("[processPacket] RSSI ");
    Serial.print(LoRa.packetRssi());
    Serial.print("   SNR ");
    Serial.println(LoRa.packetSnr());
    Serial.println("");
    
    display.drawHorizontalLine(0, 52, 128);
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0, 54, "RSSI " + (String)LoRa.packetRssi() + "dB");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(128, 54, "SNR " + (String)LoRa.packetSnr() +"dB");
    display.display();

//    LoRa.dumpRegisters(Serial);
//    Serial.println("");
}

void updateMQTT(genericPayload packetPayload, int packetType, uint32_t sourceMAC) {
  bool valueOutOfRange = false;
  
  Serial.println("[updateMQTT] Sending MQTT message at " + timeClient.getFormattedTime());
  Serial.print("[updateMQTT] Connecting to ");
  Serial.println(MQTT_Server);

// Publish Temperature, Humidity and Pressure readings

  switch (packetType)
  {
    case powerTypePacket:
    {
      float batteryVoltage = packetPayload.power.battery.voltage/1000;
      if ( (batteryVoltage < batteryVoltageFloor) or (batteryVoltage > batteryVoltageCeiling) ) {
        Serial.print("[updateMQTT] Voltage out of Range ");
        Serial.print( batteryVoltageFloor );
        Serial.print(" - ");
        Serial.print( batteryVoltageCeiling );
        Serial.print(" : ");
        Serial.println( batteryVoltage );
      } else {
        // Publish the voltage reading
        static char voltageChar[6];
        dtostrf(batteryVoltage, 6, 2, voltageChar);
        switch (sourceMAC)
        {
          case node1MAC:
          {
            client.publish("monitor/voltage", voltageChar);
            client.publish("monitor/heartbeat", "true");
            Serial.println("[updateMQTT] Node 32 Voltage message sent");
            heartbeat(node1MAC);
            break;
          }
        }
      }
      break;
    }
    case voltageTypePacket:
    {
      float sourceVoltage = packetPayload.voltage.source.voltage/1000;
      if ( (sourceVoltage < batteryVoltageFloor) or (sourceVoltage > batteryVoltageCeiling) ) {
        Serial.print("[updateMQTT] Voltage out of Range ");
        Serial.print( batteryVoltageFloor );
        Serial.print(" - ");
        Serial.print( batteryVoltageCeiling );
        Serial.print(" : ");
        Serial.println( sourceVoltage );
      } else {
        // Publish the voltage reading
        static char voltageChar[6];
        dtostrf(sourceVoltage, 6, 2, voltageChar);
        switch (sourceMAC)
        {
          case nodeAMAC:
          {
            client.publish("porch/voltage", voltageChar);
            client.publish("porch/heartbeat", "true");
            Serial.println("[updateMQTT] Porch Mini Voltage message sent");
            heartbeat(nodeAMAC);
            break;
          }
          case nodeBMAC:
          {
            client.publish("shed/voltage", voltageChar);
            client.publish("shed/heartbeat", "true");
            Serial.println("[updateMQTT] Shed Mini Voltage message sent");
            heartbeat(nodeBMAC);
            break;
          }
         case node8MAC:
          {
            client.publish("minibb/voltage", voltageChar);
            client.publish("minibb/heartbeat", "true");
            Serial.println("[updateMQTT] BB Mini Voltage message sent");
            heartbeat(node8MAC);
            break;
          }
        }
      }
      break;
    }
    case tankTypePacket:
    {
      uint8_t tankLevel = packetPayload.tank.tank.level;
      if ( (tankLevel < tankLevelFloor) or (tankLevel > tankLevelCeiling) ) {
        Serial.print("[updateMQTT] Tank Level out of Range ");
        Serial.print( tankLevelFloor );
        Serial.print(" - ");
        Serial.print( tankLevelCeiling );
        Serial.print(" : ");
        Serial.println( tankLevel );
      } else {
        // Publish the tank level reading
        static char levelChar[3];
        itoa(tankLevel, levelChar, 10);
        switch (sourceMAC)
        {
          case node5MAC:
          {
            client.publish("tank/level", levelChar);
            client.publish("tank/heartbeat", "true");
            Serial.println("[updateMQTT] Stick Tank Level message sent");
             heartbeat(node5MAC);
           break;
          }
        }
      }
      break;
    }
    case pumpTypePacket:
    {
      Serial.println("[updateMQTT] Pump detsils not currently recorded");
      break;
    }
    case weatherTypePacket:
    {
      float temperature = packetPayload.atmosphere.recorded.temperature/10;
      uint16_t pressure = packetPayload.atmosphere.recorded.pressure;
      uint16_t humidity = packetPayload.atmosphere.recorded.humidity;
      float recordedRainfall = packetPayload.rainfall.recorded.rainfall/10;
      uint16_t windDirection = packetPayload.wind.recorded.windDirection;
      uint8_t windSpeed = packetPayload.wind.recorded.windSpeed;
      
      if ( (temperature < temperatureFloor) or (temperature > temperatureCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Tempreature out of Range ");
        Serial.print( temperatureFloor );
        Serial.print(" - ");
        Serial.print( temperatureCeiling );
        Serial.print(" : ");
        Serial.println( temperature );
      }
      
      if ( (pressure < pressureFloor) or (pressure > pressureCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Pressure out of Range ");
        Serial.print( pressureFloor );
        Serial.print(" - ");
        Serial.print( pressureCeiling );
        Serial.print(" : ");
        Serial.println( pressure );
      }
      
      if ( (humidity < humidityFloor) or (humidity > humidityCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Humidity out of Range ");
        Serial.print( humidityFloor );
        Serial.print(" - ");
        Serial.print( humidityCeiling );
        Serial.print(" : ");
        Serial.println( humidity );
      }
      
      if ( (recordedRainfall < rainfallFloor) or (recordedRainfall > rainfallCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Rainfall out of Range ");
        Serial.print( rainfallFloor );
        Serial.print(" - ");
        Serial.print( rainfallCeiling );
        Serial.print(" : ");
        Serial.println( recordedRainfall );
      }
      
      if ( (windDirection < windDirectionFloor) or (windDirection > windDirectionCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Wind Direciton out of Range ");
        Serial.print( windDirectionFloor );
        Serial.print(" - ");
        Serial.print( windDirectionCeiling );
        Serial.print(" : ");
        Serial.println( windDirection );
      }
      
      if ( (windSpeed < windSpeedFloor) or (windSpeed > windSpeedCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Wind Speed out of Range ");
        Serial.print( windSpeedFloor );
        Serial.print(" - ");
        Serial.print( windSpeedCeiling );
        Serial.print(" : ");
        Serial.println( windSpeed );
      }

      if ( valueOutOfRange ) {
        Serial.println("[updateMQTT] Update aborted");   
      } else {
        // Publish sensor readings
        static char temperatureChar[6];
        dtostrf(temperature, 5, 1, temperatureChar);
        static char humidityChar[6];
        dtostrf(humidity, 5, 1, humidityChar);
        static char pressureChar[8];
        dtostrf(pressure, 7, 1, pressureChar);
        static char rainfallChar[6];
        dtostrf(recordedRainfall, 5, 1, rainfallChar);
        static char windDirectionChar[4];
        itoa(windDirection, windDirectionChar, 10);
        static char windSpeedChar[4];
        itoa(windSpeed, windSpeedChar, 10);

        switch (sourceMAC)
        {
          case node2MAC:
          {
            client.publish("rover/humidity", humidityChar);
            client.publish("rover/temperature", temperatureChar);
            client.publish("rover/pressure", pressureChar);
            client.publish("rover/rainfall", rainfallChar);
            client.publish("rover/direction", windDirectionChar);
            client.publish("rover/speed", windSpeedChar);
            client.publish("rover/heartbeat", "true");
            Serial.println("[updateMQTT] Rover BME Weather messages sent");
            heartbeat(node2MAC);
            break;
          }
          case node7MAC:
          {
            client.publish("paddock/humidity", humidityChar);
            client.publish("paddock/temperature", temperatureChar);
            client.publish("paddock/pressure", pressureChar);
            client.publish("paddock/rainfall", rainfallChar);
            client.publish("paddock/direction", windDirectionChar);
            client.publish("paddock/speed", windSpeedChar);
            client.publish("paddock/heartbeat", "true");
            Serial.println("[updateMQTT] Cell BME+ Weather messages sent");
            heartbeat(node7MAC);
            break;
          }
          default:
          {
            Serial.println("[updateMQTT] No node/topic configured");
            break;
          }
        }
      }     
      break;
    }
    case atmosphereTypePacket:
    {
      float temperature = packetPayload.atmosphere.recorded.temperature/10;
      uint16_t pressure = packetPayload.atmosphere.recorded.pressure;
      uint8_t humidity = packetPayload.atmosphere.recorded.humidity;
      
      if ( (temperature < temperatureFloor) or (temperature > temperatureCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Tempreature out of Range ");
        Serial.print( temperatureFloor );
        Serial.print(" - ");
        Serial.print( temperatureCeiling );
        Serial.print(" : ");
        Serial.println( temperature );
      }
      
      if ( (pressure < pressureFloor) or (pressure > pressureCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Pressure out of Range ");
        Serial.print( pressureFloor );
        Serial.print(" - ");
        Serial.print( pressureCeiling );
        Serial.print(" : ");
        Serial.println( pressure );
      }
      
      if ( (humidity < humidityFloor) or (humidity > humidityCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Humidity out of Range ");
        Serial.print( humidityFloor );
        Serial.print(" - ");
        Serial.print( humidityCeiling );
        Serial.print(" : ");
        Serial.println( humidity );
      }
      
      if ( valueOutOfRange ) {
        Serial.println("[updateMQTT] Update aborted");   
      } else {
        // Publish Temperature, Humidity and Pressure readings
        static char temperatureChar[6];
        dtostrf(temperature, 5, 1, temperatureChar);
        static char humidityChar[6];
        dtostrf(humidity, 5, 1, humidityChar);
        static char pressureChar[8];
        dtostrf(pressure, 7, 1, pressureChar);

        switch (sourceMAC)
        {
          case node2MAC:
          {
            client.publish("study/humidity", humidityChar);
            client.publish("study/temperature", temperatureChar);
            client.publish("study/pressure", pressureChar);
            client.publish("study/heartbeat", "true");
            Serial.println("[updateMQTT] Study BME Atmosphere messages sent");
            heartbeat(node2MAC);
            break;
          }
          case node3MAC:
          {
            client.publish("rover/humidity", humidityChar);
            client.publish("rover/temperature", temperatureChar);
            client.publish("rover/pressure", pressureChar);
            client.publish("rover/heartbeat", "true");
            Serial.println("[updateMQTT] Rover BME Atmosphere messages sent");
            heartbeat(node3MAC);
            break;
          }
          case node7MAC:
          {
            client.publish("paddock/humidity", humidityChar);
            client.publish("paddock/temperature", temperatureChar);
            client.publish("paddock/pressure", pressureChar);
            client.publish("paddock/heartbeat", "true");
            Serial.println("[updateMQTT] Cell BME Atmosphere messages sent");
            heartbeat(node7MAC);
            break;
          }
          case node8MAC:
          {
            client.publish("minibb/humidity", humidityChar);
            client.publish("minibb/temperature", temperatureChar);
            client.publish("minibb/pressure", pressureChar);
            client.publish("minibb/heartbeat", "true");
            Serial.println("[updateMQTT] BB Mini BME Atmosphere messages sent");
            heartbeat(node8MAC);
            break;
          }
           case nodeAMAC:
          {
            client.publish("porch/humidity", humidityChar);
            client.publish("porch/temperature", temperatureChar);
            client.publish("porch/pressure", pressureChar);
            client.publish("porch/heartbeat", "true");
            Serial.println("[updateMQTT] Porch Mini BME Atmosphere messages sent");
            heartbeat(nodeAMAC);
            break;
          }
           case nodeBMAC:
          {
            client.publish("shed/humidity", humidityChar);
            client.publish("shed/temperature", temperatureChar);
            client.publish("shed/pressure", pressureChar);
            client.publish("shed/heartbeat", "true");
            Serial.println("[updateMQTT] Shed Mini BME Atmosphere messages sent");
            heartbeat(nodeBMAC);
            break;
          }
          default:
          {
            Serial.println("[updateMQTT] No node/topic configured");
            break;
          }
        }
      }
      break;
    }
    case rainfallTypePacket:
    {
      float recordedRainfall = packetPayload.rainfall.recorded.rainfall/10;
      if ( (recordedRainfall < rainfallFloor) or (recordedRainfall > rainfallCeiling) ) {
        Serial.print("[updateMQTT] Rainfall out of Range ");
        Serial.print( rainfallFloor );
        Serial.print(" - ");
        Serial.print( rainfallCeiling );
        Serial.print(" : ");
        Serial.println( recordedRainfall );
      } else {
        // Publish the rainfall reading
        static char rainfallChar[6];
        dtostrf(recordedRainfall, 5, 1, rainfallChar);
        switch (sourceMAC)
        {
          case node4MAC:
          {
            client.publish("raingauge/rainfall", rainfallChar);
            Serial.println("[updateMQTT] Stick Rainfall message sent");
            client.publish("raingauge/heartbeat", "true");
            heartbeat(node4MAC);
            break;
          }
        }
      }
      break;
    }
    case windTypePacket:
    {
      uint16_t windDirection = packetPayload.wind.recorded.windDirection;
      uint8_t windSpeed = packetPayload.wind.recorded.windSpeed;
      
      if ( (windDirection < windDirectionFloor) or (windDirection > windDirectionCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Wind Direciton out of Range ");
        Serial.print( windDirectionFloor );
        Serial.print(" - ");
        Serial.print( windDirectionCeiling );
        Serial.print(" : ");
        Serial.println( windDirection );
      }
      
      if ( (windSpeed < windSpeedFloor) or (windSpeed > windSpeedCeiling) ) {
        valueOutOfRange = true;
        Serial.print("[updateMQTT] Wind Speed out of Range ");
        Serial.print( windSpeedFloor );
        Serial.print(" - ");
        Serial.print( windSpeedCeiling );
        Serial.print(" : ");
        Serial.println( windSpeed );
      }

      if ( valueOutOfRange ) {
        Serial.println("[updateMQTT] Update aborted");   
      } else {
        // Publish wind direction and speed readings
        static char windDirectionChar[4];
        itoa(windDirection, windDirectionChar, 10);
        static char windSpeedChar[4];
        itoa(windSpeed, windSpeedChar, 10);

        switch (sourceMAC)
        {
          case node6MAC:
          {
            client.publish("house/direction", windDirectionChar);
            client.publish("house/speed", windSpeedChar);
            client.publish("house/heartbeat", "true");
            Serial.println("[updateMQTT] Statuc Cube Wind messages sent");
            heartbeat(node6MAC);
            break;
          }
          case node7MAC:
          {
            client.publish("paddock/direction", windDirectionChar);
            client.publish("paddock/speed", windSpeedChar);
            client.publish("paddock/heartbeat", "true");
            Serial.println("[updateMQTT] Cube BME+ Wind messages sent");
            heartbeat(node7MAC);
            break;
          }
        }
      }
      break;
    }
    case sprinklerTypePacket:
    {
      Serial.println("[updateMQTT] Sprinkler details not currently recorded");
      break;
    }
//    default:
//      Nothing here yet
  }
}
void heartbeat( uint32_t nodeID ) {
  switch (nodeID)
  {
    case node1MAC :
    {
      node1Timer = 0;
      break;
    }
    case node2MAC :
    {
      node2Timer = 0;
      break;
    }
    case node3MAC :
    {
      node3Timer = 0;
      break;
    }
    case node4MAC :
    {
      node4Timer = 0;
      break;
    }
    case node5MAC :
    {
      node5Timer = 0;
      break;
    }
    case node6MAC :
    {
      node6Timer = 0;
      break;
    }
    case node7MAC :
    {
      node7Timer = 0;
      break;
    }
    case node8MAC :
    {
      node8Timer = 0;
      break;
    }
    case node9MAC :
    {
      node9Timer = 0;
      break;
    }
    case nodeAMAC :
    {
      nodeATimer = 0;
      break;
    }
    case nodeBMAC :
    {
      nodeBTimer = 0;
      break;
    }
  }
}

void recordHeartbeat() {
  node1Timer = node1Timer + 1;
  node2Timer = node2Timer + 1;
  node3Timer = node3Timer + 1;
  node4Timer = node4Timer + 1;
  node5Timer = node5Timer + 1;
  node6Timer = node6Timer + 1;
  node7Timer = node7Timer + 1;
  node8Timer = node8Timer + 1;
  node9Timer = node9Timer + 1;
  nodeATimer = nodeATimer + 1;
  nodeBTimer = nodeBTimer + 1;
}

void healthCheck() {
  if ( node1Timer > deadBeat ) {
    client.publish("monitor/heartbeat", "false");
    Serial.println("[healthCheck] Node 1 (Power WiFi LoRa 32) heart failure");
  }
  if ( node2Timer > deadBeat ) {
    client.publish("study/heartbeat", "false");
    Serial.println("[healthCheck] Node 2 (Rover ESP32S) heart failure");
  }
  if ( node3Timer > deadBeat ) {
    client.publish("rover/heartbeat", "false");
    Serial.println("[healthCheck] Node 3 (Study ESP32S) heart failure");
  }
  if ( node4Timer > deadBeat ) {
    client.publish("raingauge/heartbeat", "false");
    Serial.println("[healthCheck] Node 4 (Rain Stick) heart failure");
  }
  if ( node5Timer > deadBeat ) {
    client.publish("tank/heartbeat", "false");
    Serial.println("[healthCheck] Node 5 (Tank Stick) heart failure");
  }
  if ( node6Timer > deadBeat ) {
    client.publish("house/heartbeat", "false");
    Serial.println("[healthCheck] Node 6 (House Cube Wind) heart failure");
  }
  if ( node7Timer > deadBeat ) {
    client.publish("paddock/heartbeat","false");
    Serial.println("[healthCheck] Node 7 (Paddock Cube Weather) heart failure");
  }
  if ( node8Timer > deadBeat ) {
    client.publish("bb/heartbeat", "false");
    Serial.println("[healthCheck] Node 8 (Breadboard Mini BME) heart failure");
  }
  if ( node9Timer > deadBeat ) {
    client.publish("wijnand2/heartbeat", "false");
    Serial.println("[healthCheck] Node 9 (Wij-2 Mini BME) heart failure");
  }
  if ( nodeATimer > deadBeat ) {
    client.publish("porch/heartbeat", "false");
    Serial.println("[healthCheck] Node A (Porch Mini BME) heart failure");
  }
  if ( nodeBTimer > deadBeat ) {
    client.publish("shed/heartbeat", "false");
    Serial.println("[healthCheck] Node B (Shed Mini BME) heart failure");
  }
}
